{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "f1057e2b-7ddb-4cf2-ab01-f18b6eeb1fc2",
   "metadata": {},
   "source": [
    "# Exercise: Combining bayesian linear models and neural networks"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "299a34f9-b805-4ebd-9f54-60bdd31be0c0",
   "metadata": {
    "tags": [
     "disable-execution-page"
    ]
   },
   "outputs": [],
   "source": [
    "import math\n",
    "import torch\n",
    "import torch.nn as nn\n",
    "import torch.nn.functional as F\n",
    "from torch.utils.data import DataLoader\n",
    "from matplotlib import pyplot as plt\n",
    "from matplotlib.pyplot import Slider\n",
    "\n",
    "%matplotlib widget\n",
    "from cycler import cycler\n",
    "import seaborn as sns\n",
    "\n",
    "# Set the color scheme\n",
    "sns.set_theme()\n",
    "colors = [\n",
    "    \"#0076C2\",\n",
    "    \"#EC6842\",\n",
    "    \"#A50034\",\n",
    "    \"#009B77\",\n",
    "    \"#FFB81C\",\n",
    "    \"#E03C31\",\n",
    "    \"#6CC24A\",\n",
    "    \"#EF60A3\",\n",
    "    \"#0C2340\",\n",
    "    \"#00B8C8\",\n",
    "    \"#6F1D77\",\n",
    "]\n",
    "plt.rcParams[\"axes.prop_cycle\"] = cycler(color=colors)\n",
    "torch.set_default_dtype(torch.float64)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "578dbd11",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Class that normalizes data to follow Normal(0, 1) distribution.\n",
    "class normUnitvar:\n",
    "    def __init__(self, fullDataset):\n",
    "        self.normmean = fullDataset.mean(axis=0)\n",
    "        self.normstd = fullDataset.std(axis=0)\n",
    "\n",
    "    def normalize(self, data):\n",
    "        return (data - self.normmean) / self.normstd\n",
    "\n",
    "    def denormalize(self, data):\n",
    "        return data * self.normstd + self.normmean"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "1fd4276f-b170-4ab0-b5b4-bf278a38924e",
   "metadata": {
    "tags": []
   },
   "source": [
    "## Introduction\n",
    "In this exercise notebook, we will combine a bayesian linear model and a neural network to create a very flexible Bayesian model.\n",
    "\n",
    "In a standard bayesian linear model the basis functions are pre-defined.\n",
    "Here, we use the flexibility of neural networks instead to find a good set of basis functions from data.\n",
    "Take a look of the proposed model in this schematic figure:\n",
    "\n",
    "<div style=\"text-align:center\">\n",
    "    <img src=\"https://surfdrive.surf.nl/files/index.php/s/5mQ8hTliDTbCFES/download\" width=\"350\" float=\"center\"/>\n",
    "</div>\n",
    "\n",
    "The model is created by first training a neural network to predict $y$ based on $x$. Then, keeping the network fixed, we replace the output layer of the network with the bayesian linear model, which uses the output of the last hidden layer as input.\n",
    "\n",
    "After creating this model, we will use emperical bayes as discussed in the lecture to fit the hyperparameters."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "103e0416-1d18-4964-bdf9-00935c62e673",
   "metadata": {},
   "source": [
    "First, we will create a simple dataset.\n",
    "Our data is generated by `f_data` which takes the ground truth but corrupts it with i.i.d. Gaussian noise, imitating a realistic setting where we do not have perfect measurements.\n",
    "\n",
    "After generating our data, we standardize it and wrap it with a dataloader."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0297490b-26f1-49e4-a82b-5a49edd5ea1e",
   "metadata": {},
   "outputs": [],
   "source": [
    "# The true function relating t to x\n",
    "def f_truth(x):\n",
    "    # Return a sine\n",
    "    return torch.sin(x)\n",
    "\n",
    "\n",
    "# The data is generated from the ground truth with i.i.d. gaussian noise\n",
    "def f_data(x, beta_true, rng):\n",
    "    # Generate N noisy observations (1 at each location)\n",
    "    t = f_truth(x) + torch.normal(\n",
    "        mean=0, std=math.sqrt(1 / beta_true), size=x.shape, generator=rng\n",
    "    )\n",
    "\n",
    "    # Return the observations\n",
    "    return t"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "923c8311",
   "metadata": {},
   "outputs": [],
   "source": [
    "x = torch.arange(0, 10, 0.2).reshape(-1, 1)  # Reshape necessary for network prediction\n",
    "\n",
    "# We manually remove data to create some gaps in the data.\n",
    "x = torch.cat((x[:10], x[18:29], x[43:]), dim=0)\n",
    "\n",
    "rng = torch.manual_seed(1)\n",
    "beta_true = 50\n",
    "\n",
    "t = f_data(x, beta_true, rng)\n",
    "\n",
    "fig, ax = plt.subplots(1, 1, figsize=(6, 4))\n",
    "ax.scatter(x, t, marker=\"x\")\n",
    "ax.set_xlabel(\"x\")\n",
    "ax.set_ylabel(\"y\")\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c7585035",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create normalizers and normalize the data\n",
    "x_normalizer = normUnitvar(x)\n",
    "t_normalizer = normUnitvar(t)\n",
    "\n",
    "x_norm = x_normalizer.normalize(x)\n",
    "t_norm = t_normalizer.normalize(t)\n",
    "\n",
    "# We create a simple dataset with only training data and a single batch.\n",
    "training_dataset = torch.utils.data.TensorDataset(x_norm, t_norm)\n",
    "\n",
    "train_loader = DataLoader(training_dataset, batch_size=len(x), shuffle=True)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "cc9735d1",
   "metadata": {},
   "source": [
    "## Exercise: Defining the neural network\n",
    "The first step in making the model is to create the neural network. The skelleton of the model is in place, but the following tasks still need to be done:\n",
    "- Define a neural network. You can make it any size you want, but it should have at least 1 hidden layer.\n",
    "- Create an additional function  `learned_phi`, which uses the same layers as the forward function but returns the output of the last hidden layer, with an additional term for the bias. This output will act as the basis for the Bayesian model."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "adb0f84f",
   "metadata": {},
   "outputs": [],
   "source": [
    "class Net(nn.Module):\n",
    "    def __init__(self):\n",
    "        super(Net, self).__init__()\n",
    "        # ---------------------- student exercise --------------------------------- #\n",
    "        # YOUR CODE HERE\n",
    "        # ---------------------- student exercise --------------------------------- #\n",
    "\n",
    "    def forward(self, x):\n",
    "        # ---------------------- student exercise --------------------------------- #\n",
    "        # YOUR CODE HERE\n",
    "        # ---------------------- student exercise --------------------------------- #\n",
    "        return out\n",
    "\n",
    "    def learned_phi(self, x):\n",
    "        # ---------------------- student exercise --------------------------------- #\n",
    "        # YOUR CODE HERE\n",
    "        # ---------------------- student exercise --------------------------------- #\n",
    "        # Add bias term        \n",
    "        phi = torch.cat((torch.ones(x.shape[0], 1), x), dim=1)\n",
    "        return phi\n",
    "\n",
    "\n",
    "net = Net()\n",
    "print(net)\n",
    "\n",
    "# Define the optimizer\n",
    "optimizer = torch.optim.Adam(net.parameters(), lr=0.1)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9c57f3f6",
   "metadata": {},
   "source": [
    "### Exercise: Training loop\n",
    "Your next task is to train the neural network. which can be broken down into the following steps:\n",
    "- Implement a neural network training loop. Think about which parts are necessary when optimizing a Bayesian model.\n",
    "- Have the function return the trained network and the loss curve.\n",
    "- Call your function to train the model"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "605f78c6",
   "metadata": {},
   "outputs": [],
   "source": [
    "def train_nn(network, train_loader, epochs, optimizer):\n",
    "    \"\"\"\n",
    "    Class that trains a network\n",
    "    :param network: torch.nn class with forward function\n",
    "    :param train_loader: DataLoader object containing all training batches\n",
    "    :param epochs: Int, number of epochs to train the model for\n",
    "    :param optimizer: The optimizer used (e.g. ADAM/SGD)\n",
    "    :return: [Trained network,\n",
    "              Tensor of MSE losses per epoch]\n",
    "    \"\"\"\n",
    "\n",
    "    # ---------------------- student exercise --------------------------------- #\n",
    "    # YOUR CODE HERE\n",
    "    # ---------------------- student exercise --------------------------------- #\n",
    "\n",
    "    return network, train_losses"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0a424123",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Train the network\n",
    "# Call your neural network training function, store the network as \"net\" and train_losses as \"losses\"\n",
    "# ---------------------- student exercise --------------------------------- #\n",
    "# YOUR CODE HERE\n",
    "# ---------------------- student exercise --------------------------------- #\n",
    "\n",
    "# Plotting the loss curve\n",
    "fig, ax = plt.subplots(1, 1, figsize=(6, 4))\n",
    "ax.plot(torch.arange(len(losses)), losses, label=\"Training\")\n",
    "ax.legend()\n",
    "ax.set_xlabel(f\"Epochs\")\n",
    "ax.set_ylabel(f\"MSE\")\n",
    "ax.set_title(f\"Loss curve\")\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e0019629",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Plot the result\n",
    "x, t = next(iter(train_loader))  # Get data\n",
    "\n",
    "# Create prediction array\n",
    "x_pred = torch.arange(-2, 12, 0.1).reshape(\n",
    "    -1, 1\n",
    ")  # Reshape necessary for network prediction\n",
    "\n",
    "x_pred_norm = x_normalizer.normalize(x_pred)  # normalized data\n",
    "\n",
    "# Predictions\n",
    "pred = net(x_pred_norm).detach()  # Make prediction\n",
    "\n",
    "# Denormalize\n",
    "x_denorm = x_normalizer.denormalize(x)\n",
    "t_denorm = t_normalizer.denormalize(t)\n",
    "pred_denorm = t_normalizer.denormalize(pred)\n",
    "\n",
    "# Scatter plot for shuffled data\n",
    "fig, ax = plt.subplots(1, 1, figsize=(6, 4))\n",
    "ax.scatter(x_denorm, t_denorm, label=\"True\", color=\"C0\", marker=\"x\")\n",
    "ax.plot(x_pred, pred_denorm, label=\"Prediction\", color=\"C1\")\n",
    "ax.set_xlabel(\"x\")\n",
    "ax.set_ylabel(\"y\")\n",
    "ax.legend()\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "722847c3265fb279",
   "metadata": {},
   "source": [
    "If you've succesfully trained your model, you should see a reasonable overlap between the true values and model predictions.\n",
    "\n",
    "## Bayesian Linear Model\n",
    "With the neural network trained, we can turn our attention to the Bayesian Linear Model. \n",
    "We will create a class here that computes the mean and covariance of the posterior.\n",
    "\n",
    "Recall from the lecture that the posterior mean is defined as\n",
    "\n",
    "$$\n",
    "\\mathbf{m} = \\beta\\mathbf{S}\\boldsymbol{\\Phi}^\\mathrm{T}\\mathbf{t},\n",
    "$$\n",
    "\n",
    "and its covariance as\n",
    "\n",
    "$$\n",
    "\\mathbf{S}^{-1} = \\alpha\\mathbf{I} + \\beta\\boldsymbol{\\Phi}^\\mathrm{T}\\boldsymbol{\\Phi}.\n",
    "$$\n",
    "\n",
    "In addition, we create a function which returns the output distribution — again defined by its mean and covariance — for a new input: \n",
    "\n",
    "$$\n",
    "p(\\hat{t}\\vert\\hat{\\mathbf{x}},\\beta) =\n",
    "\\displaystyle\\mathcal{N}\\left(\n",
    "\\hat{t}\\vert\n",
    "\\mathbf{m}^\\mathrm{T}\\boldsymbol{\\phi}(\\hat{\\mathbf{x}}),\n",
    "\\frac{1}{\\beta} + \\boldsymbol{\\phi}(\\hat{\\mathbf{x}})^\\mathrm{T}\\mathbf{S}\\boldsymbol{\\phi}(\\hat{\\mathbf{x}})\n",
    "\\right).\n",
    "$$\n",
    "\n",
    "The set of basis function is passed as an input, allowing it to handle arbitrary basis functions, such as the ones learned by the neural network defined earlier."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "01c959ab",
   "metadata": {},
   "outputs": [],
   "source": [
    "class bayesmodel:\n",
    "    def __init__(self, basisfunc, x, t, alpha, beta):\n",
    "        \"\"\"\n",
    "        Class for a general Bayesian Linear Model\n",
    "        :param basisfunc: The basis function to be used\n",
    "        :param x: Dataset inputs to build the model with\n",
    "        :param t: Dataset targets to build the model with\n",
    "        :param alpha: prior precision\n",
    "        :param beta: noise precision\n",
    "        \"\"\"\n",
    "        with torch.no_grad():\n",
    "            self.phi = basisfunc\n",
    "            # Initial values\n",
    "            self.alpha = alpha\n",
    "            self.beta = beta\n",
    "\n",
    "            # Compute m_N & S_N\n",
    "            self.phi_x = self.phi(x)\n",
    "            phi_x_t = torch.transpose(self.phi_x, 0, 1)\n",
    "\n",
    "            self.cov_inverse = self.alpha * torch.eye(\n",
    "                self.phi_x.shape[1]\n",
    "            ) + self.beta * torch.matmul(phi_x_t, self.phi_x)\n",
    "            self.covariance = torch.inverse(self.cov_inverse)  # Bishop 3.54\n",
    "            self.mean = self.beta * torch.matmul(\n",
    "                self.covariance, torch.matmul(phi_x_t, t)\n",
    "            )  # Bishop 3.53\n",
    "\n",
    "    def pred(self, x_new):\n",
    "        \"\"\"\n",
    "        :param x_new: Inputs for which to compute the mean and std.\n",
    "        :return: [mean, std] for all x_new\n",
    "        \"\"\"\n",
    "        with torch.no_grad():\n",
    "            # Make new predictions\n",
    "            phi_x_new = self.phi(x_new)\n",
    "            phi_x_new_t = torch.transpose(phi_x_new, 0, 1)\n",
    "            pred_m = torch.matmul(phi_x_new, self.mean)\n",
    "\n",
    "            S_N = (1 / self.beta) * torch.eye(phi_x_new.shape[0]) + torch.matmul(\n",
    "                phi_x_new, torch.matmul(self.covariance, phi_x_new_t)\n",
    "            )  # Biship 3.59\n",
    "\n",
    "            pred_var = torch.unsqueeze(torch.diag(S_N), 1)\n",
    "\n",
    "        return pred_m, pred_var"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "38cbedac6524210c",
   "metadata": {},
   "source": [
    "Our Bayesian Linear Model is complete, and we can use it to make predictions. \n",
    "The final task for us is to find which values of $\\alpha$ and $\\beta$ are appropriate. \n",
    "Before we do that, let's see our model in action assuming that we know these hyperparameters.\n",
    "\n",
    "## Assuming known hyperparameters\n",
    "Let's test our model to see if it can make accurate predictions if we a priori fix the hyperparameters $\\alpha$ and $\\beta$.\n",
    "Try different values for $\\alpha$ and $\\beta$. What happens if you set either to very large or very small values?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "6e0ea9ec-d81f-4bab-a1e1-42255e062542",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Load data for making bayesian model (same as NN training data)\n",
    "x, t = next(iter(train_loader))\n",
    "\n",
    "# To show the outcome, we use fixed values for alpha and beta here:\n",
    "alpha = 1\n",
    "beta = 50\n",
    "\n",
    "# We pass the hidden network layer as basis function\n",
    "bayesNet = bayesmodel(net.learned_phi, x, t, alpha, beta)\n",
    "\n",
    "# Get truth\n",
    "target = f_truth(x_pred)\n",
    "\n",
    "# Predictions\n",
    "pred_mean, pred_var = bayesNet.pred(x_pred_norm)\n",
    "\n",
    "# Denormalize outputs\n",
    "pred = t_normalizer.denormalize(pred_mean.squeeze())\n",
    "variance = t_normalizer.denormalize(pred_var.squeeze())\n",
    "t_data = t_normalizer.denormalize(t)\n",
    "\n",
    "# Plotting\n",
    "fig, ax = plt.subplots(1, 1, figsize=(6, 4))\n",
    "ax.plot(x_pred.squeeze(), target, label=\"True\", color=\"black\")\n",
    "pred_plt = ax.plot(x_pred.squeeze(), pred, label=\"Prediction\", color=\"C1\")\n",
    "if variance is not None:\n",
    "    std = torch.sqrt(variance)\n",
    "    ax.fill_between(\n",
    "        x_pred.squeeze(),\n",
    "        pred - 1.96 * std,\n",
    "        pred + 1.96 * std,\n",
    "        color=\"C1\",\n",
    "        alpha=0.3,\n",
    "        linewidth=0,\n",
    "        label=\"95% conf. interval\",\n",
    "    )\n",
    "ax.scatter(x_normalizer.denormalize(x), t_data, label=\"Data\", color=\"C0\", marker=\"x\")\n",
    "ax.legend()\n",
    "plt.title(f\"alpha={alpha:.2f}, beta={beta:.2f}\")\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "23234451f878935",
   "metadata": {},
   "source": [
    "## Dealing with unknown hyperparameters\n",
    "In most realistic cases, $\\alpha$ and $\\beta$ are not known beforehand.\n",
    "Ideally, we would introduce priors for $\\alpha$ and $\\beta$, and make predictions by marginalizing with respect to these hyperparameters as well as with respect to the parameters w.\n",
    "However, the complete marginalization over all of these variables is analytically intractable.\n",
    "As discussed in the lecture, we can use an approximation, where we first integrate over the parameters w, and then set the hyperparameters $\\alpha$ and $\\beta$ by maximizing the marginal likelihood.\n",
    "This is known as Empirical Bayes or the evidence approximation. \n",
    "\n",
    "\n",
    "\n",
    "### Exercise:\n",
    "As discussed in the lecture, the marginal likelihood can be computed as:\n",
    "\n",
    "$$\n",
    "\\ln p(\\mathbf{t}\\vert\\alpha,\\beta) = \\displaystyle\\frac{N}{2}\\ln\\beta\n",
    "-\\frac{\\beta}{2}\\sum_{n=1}^{N}\\left[t_n-\\mathbf{m}_N^\\mathrm{T}\\boldsymbol{\\phi}(\\mathbf{x}_n)\\right]^2\n",
    "-\\frac{\\alpha}{2}\\mathbf{m}_N^\\mathrm{T}\\mathbf{m}_N\n",
    "+ \\frac{M}{2}\\ln\\alpha - \\frac{1}{2}\\ln\\vert\\mathbf{S}_N\\vert - \\frac{N}{2}\\ln(2\\pi),\n",
    "$$\n",
    "\n",
    "where $N$ is the number of samples, $M$ the number of basis functions, and $\\mathbf{m}_N$ and $\\mathbf{S}_N$ come from the posterior of the weights.\n",
    "\n",
    "It is your task to finish the implementation of this function below.\n",
    "\n",
    "(Note: `M = net.hid_n + 1` will probably raise an error. Change `hid_n` to your own variable name definition for the number of hidden nodes in `class Net` defined in the first exercise.)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "779c80e4e98df0e1",
   "metadata": {},
   "outputs": [],
   "source": [
    "def logmarginal(alpha, beta, x, t):\n",
    "    # Create a new model\n",
    "    model = bayesmodel(net.learned_phi, x, t, alpha, beta)\n",
    "\n",
    "    # We need to compute the determinant of the matrix of second derivatives of the error funcion (the Hessian) (see Biship 3.81).\n",
    "    # This is equivalent to S_N:\n",
    "    A = model.cov_inverse\n",
    "    detA = torch.linalg.det(A)\n",
    "\n",
    "    M = net.hid_n + 1  # Number of 'basis functions' (hidden nodes + bias)\n",
    "    N = x.shape[0]  # Number of samples\n",
    "\n",
    "    # Compute the error\n",
    "    diff = (t - model.phi_x @ model.mean).flatten()\n",
    "\n",
    "    model_m = model.mean.flatten()\n",
    "\n",
    "    # ---------------------- student exercise --------------------------------- #\n",
    "    # YOUR CODE HERE\n",
    "    # ---------------------- student exercise --------------------------------- #\n",
    "\n",
    "    return log_marg_like"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "187d664b9c9475e5",
   "metadata": {},
   "source": [
    "With the marginal likelihood implemented, we can optimize our hyperparameters. \n",
    "We take advantage of automatic differentiation, allowing us to use a gradient-based optimizer BFGS.\n",
    "The BFGS optimizer finds the values for $\\alpha$ and $\\beta$ that maximize the model evidence. \n",
    "\n",
    "Finally, we plot the results for each iteration in an interactive plot."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "59cb9511f9a4ce56",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Initialize logalpha and logbeta, track the gradients.\n",
    "logalpha = torch.tensor([0.0], requires_grad=True)  # Log(1) = 0\n",
    "logbeta = torch.tensor([0.0], requires_grad=True)\n",
    "\n",
    "# We use the bfgs optimizer.\n",
    "bfgs = torch.optim.LBFGS((logalpha, logbeta), max_iter=20)\n",
    "\n",
    "alphas = []\n",
    "betas = []\n",
    "\n",
    "\n",
    "def update():  # This function is called by the optimizer, it cannot have any arguments, therefore we use the global variables.\n",
    "    bfgs.zero_grad()\n",
    "    loss = -logmarginal(torch.exp(logalpha), torch.exp(logbeta), x, t)\n",
    "    loss.backward()\n",
    "\n",
    "    alphas.append(torch.exp(logalpha).detach()[0])\n",
    "    betas.append(torch.exp(logbeta).detach()[0])\n",
    "    return loss\n",
    "\n",
    "\n",
    "# We perform a single step. In this step, multiple iterations of the optimizer are performed.\n",
    "bfgs.step(update)\n",
    "\n",
    "print(\"BFGS optimal result: alpha\", alphas[-1], \"beta\", betas[-1])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "94251c198adf7e1e",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Initialize a new model with the alpha and beta\n",
    "bayesNet = bayesmodel(net.learned_phi, x, t, alphas[0], betas[0])\n",
    "# Predictions\n",
    "pred_mean, pred_var = bayesNet.pred(x_pred_norm)\n",
    "\n",
    "fig, ax = plt.subplots(1, 1, figsize=(6, 4))\n",
    "fig.subplots_adjust(bottom=0.25)\n",
    "\n",
    "# Denormalize outputs\n",
    "pred = t_normalizer.denormalize(pred_mean.squeeze())\n",
    "variance = t_normalizer.denormalize(pred_var.squeeze())\n",
    "std = torch.sqrt(variance)\n",
    "t_data = t_normalizer.denormalize(t)\n",
    "\n",
    "# Plotting\n",
    "(truth_l,) = ax.plot(x_pred.squeeze(), target, label=\"True\", color=\"black\")\n",
    "\n",
    "(pred_l,) = ax.plot(x_pred.squeeze(), pred, label=\"Prediction\", color=\"C1\")\n",
    "var_l = ax.fill_between(\n",
    "    x_pred.squeeze(),\n",
    "    pred - 2 * std,\n",
    "    pred + 2 * std,\n",
    "    color=\"C1\",\n",
    "    alpha=0.3,\n",
    "    linewidth=0,\n",
    "    label=\"95% conf. interval\",\n",
    ")\n",
    "\n",
    "ax.scatter(x_normalizer.denormalize(x), t_data, label=\"Data\", color=\"C0\", marker=\"x\")\n",
    "ax.legend(loc=\"lower left\")\n",
    "ax.set_title(f\"Iteration {0}, alpha={alphas[0].item():.2f}, beta={betas[0].item():.2f}\")\n",
    "\n",
    "\n",
    "def update_plot(val):\n",
    "    val = int(val)\n",
    "    # Initialize a new model with the alpha and beta\n",
    "    bayesNet = bayesmodel(net.learned_phi, x, t, alphas[val], betas[val])\n",
    "\n",
    "    # Predictions\n",
    "    pred_mean, pred_var = bayesNet.pred(x_pred_norm)\n",
    "\n",
    "    # Denormalize outputs\n",
    "    pred = t_normalizer.denormalize(pred_mean.squeeze())\n",
    "    variance = t_normalizer.denormalize(pred_var.squeeze())\n",
    "    std = torch.sqrt(variance)\n",
    "\n",
    "    # Adjust plot\n",
    "    pred_l.set_ydata(pred)\n",
    "\n",
    "    global var_l\n",
    "    var_l.remove()\n",
    "    var_l = ax.fill_between(\n",
    "        x_pred.squeeze(),\n",
    "        pred - 1.96 * std,\n",
    "        pred + 1.96 * std,\n",
    "        color=\"C1\",\n",
    "        alpha=0.3,\n",
    "        linewidth=0,\n",
    "        label=\"95% conf. interval\",\n",
    "    )\n",
    "\n",
    "    ax.set_title(\n",
    "        f\"Iteration {val}, alpha={alphas[val].item():.2f}, beta={betas[val].item():.2f}\"\n",
    "    )\n",
    "    ax.set_ylim([-2, 2])\n",
    "\n",
    "    fig.canvas.draw_idle()\n",
    "\n",
    "\n",
    "# Make a horizontal slider to control the frequency.\n",
    "axiter = fig.add_axes([0.25, 0.1, 0.65, 0.03])\n",
    "it_slider = Slider(\n",
    "    ax=axiter,\n",
    "    label=\"BFGS iteration\",\n",
    "    valmin=0,\n",
    "    valmax=len(betas),\n",
    "    valinit=0,\n",
    "    valfmt=\"%0.0f\",\n",
    ")\n",
    "\n",
    "# register the update function with each slider\n",
    "it_slider.on_changed(update_plot)\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "94e6de0066038c44",
   "metadata": {},
   "source": [
    "Based on the outputs for the various BFGS iterations, how do $\\alpha$ and $\\beta$ influence the mean and variance of the posterior predictive distribution?\n",
    "\n",
    "## Reflection:\n",
    "\n",
    "With the code complete, it should be little work to try different networks. \n",
    "Try different network sizes and see how this affects the results.\n",
    "\n",
    "The main takeaway from this exercise should be a good conceptual understanding of how this model works.\n",
    "Answering the following questions can help you develop this understanding:\n",
    "- What happens if you use a neural network without nonlinear activation functions as a basis?\n",
    "- What happens if you use a very flexible neural network?\n",
    "- We did not use a validation set here. Why are we still not overfitting? "
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.17"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
